/*
 #
 #  File        : distance_saito.h
 #
 #  Description : CImg plugin that implements the computation of the Euclidean distance
 #                using Saito/Hirata algorithm, as described in :
 #
 #  [ST94] T. Saito and J. I. Toriwaki. New algorithms for Euclidean distance
 #         transformations of an n-dimensional digitized picture with applications.
 #         Pattern Recognition, 27:1551-1565, 1994.
 #
 #  [Hir96] T. Hirata. A unified linear-time algorithm for computing distance maps.
 #          Information Processing Letters, 58(3):129-133, May 1996.
 #
 #  [MRH00] A. Meijster, J.B.T.M. Roerdink, and W. H. Hesselink.
 #          A general algorithm for computing distance transforms in linear time.
 #          In Mathematical Morphology and its Applications to Image and Signal Processing,
 #          pages 331-340. Kluwer, 2000.
 #
 #  Copyright   : Francois-Xavier Dupe for the CImg portage
 #                ( http://www.greyc.ensicaen.fr/~fdupe/ )
 #                David Coeurjolly for the SEDT code
 #                ( http://liris.cnrs.fr/~dcoeurjo/ )
 #
 #  This software is governed by the CeCILL  license under French law and
 #  abiding by the rules of distribution of free software.  You can  use,
 #  modify and/ or redistribute the software under the terms of the CeCILL
 #  license as circulated by CEA, CNRS and INRIA at the following URL
 #  "http://www.cecill.info".
 #
 #  As a counterpart to the access to the source code and  rights to copy,
 #  modify and redistribute granted by the license, users are provided only
 #  with a limited warranty  and the software's author,  the holder of the
 #  economic rights,  and the successive licensors  have only  limited
 #  liability.
 #
 #  In this respect, the user's attention is drawn to the risks associated
 #  with loading,  using,  modifying and/or developing or reproducing the
 #  software by the user in light of its specific status of free software,
 #  that may mean  that it is complicated to manipulate,        and  that  also
 #  therefore means  that it is reserved for developers  and  experienced
 #  professionals having in-depth computer knowledge. Users are therefore
 #  encouraged to load and test the software's suitability as regards their
 #  requirements in conditions enabling the security of their systems and/or
 #  data to be ensured and,  more generally, to use and operate it in the
 #  same conditions as regards security.
 #
 #  The fact that you are presently reading this means that you have had
 #  knowledge of the CeCILL license and that you accept its terms.
 #
*/

#ifndef cimg_plugin_distance_saito
#define cimg_plugin_distance_saito

// Operators for handling specific algorithm computations
struct _saito {
  static unsigned long INFTY() {
    const unsigned long val = (unsigned long)cimg::type<long>::max();
    return val;
  }
  static unsigned long sum(const unsigned long x, const unsigned long y) {
    if ((x==INFTY()) || (y==INFTY())) return INFTY();
    return x+y;
  }
  static unsigned long prod(const unsigned long x, const unsigned long y) {
    if ((x==INFTY()) || (y==INFTY())) return INFTY();
    return x*y;
  }
  static unsigned long opp(const unsigned long x) {
    if (x==INFTY()) return INFTY();
    return -x;
  }
  static unsigned long intdivint(const unsigned long x, const unsigned long y) {
    if (x==INFTY()) return INFTY();
    if (!y) return INFTY();
    return x/y;
  }
  static unsigned long F(const unsigned long x, const unsigned long i, const unsigned long gi2, const unsigned long a) {
    return sum(a*(x-i)*a*(x-i),gi2);
  }
  static unsigned long SEP(const unsigned long i, const unsigned long u, const unsigned long gi2, const unsigned int long gu2,
                           const unsigned long a) {
    return intdivint(sum(sum((unsigned long)(a*a*u*u-a*a*i*i),gu2),opp(gi2)),2*a*a*(u-i));
  }
};

/**
 * Compute the 3D squared euclidean distance using Saito Algorithm
 * Can be use on 2D and 3D images
 * @param sY the sample size along the y-axis
 * @param sZ the sample size along the z-axis
 * @return the image of the distance
 */
CImg<typename cimg::last<T,unsigned long>::type> get_distance_saito(int sY = 1, int sZ = 1) const {
  typedef typename cimg::last<T,unsigned long>::type t_ulong;
  CImg<t_ulong> distance(*this), tmp (distance);
  distance.fill(0); // Init to zero

  // SDT along the x-direction
  cimg_forYZ(*this,y,z) {
    if (!(*this)(0,y,z)) distance(0,y,z) = 0; else distance(0,y,z) = _saito::INFTY();

    // Forward scan
    for (int x=1; x<dimx(); ++x) if (!(*this)(x,y,z)) distance(x,y,z) = 0; else distance(x,y,z) = _saito::sum(1,distance(x-1,y,z));

    // Backward scan
    for (int x=dimx()-2; x>=0; --x) if (distance(x+1,y,z)<distance(x,y,z)) distance(x,y,z) = _saito::sum(1,distance(x+1,y,z));
  }

  tmp = distance; // Save

  // SDT along the y-distance
  CImg<t_ulong> s(dimy(),1,1,1), t(s); // Center of the upper envelope parabolas and Separating index between 2 upper envelope parabolas
  unsigned long w = 0;
  long q = 0;

  cimg_forXZ(*this,x,z) {
    q = s(0) = t(0) = 0;

    // Forward Scan
    for (int u=1; u<dimy(); ++u) {
      while ((q>=0) && (_saito::F(t(q),s(q),_saito::prod(tmp(x,s(q),z),tmp(x,s(q),z)),sY)>_saito::F(t(q),u,_saito::prod(tmp(x,u,z),tmp(x,u,z)),sY)))
        --q;
      if (q<0) { q = 0; s(0) = u; }
      else {
        w = 1 + _saito::SEP(s(q),u,_saito::prod(tmp(x,s(q),z),tmp(x,s(q),z)),_saito::prod(tmp(x,u,z),tmp(x,u,z)),sY);
        if (w<height) { s(++q) = u; t(q) = w; }
      }
    }

    // Backward Scan
    for (int u=dimy()-1; u>=0; --u) {
      distance(x,u,z) = _saito::F(u,s(q),_saito::prod(tmp(x,s(q),z),tmp(x,s(q),z)),sY);
      if (t(q)==(unsigned long)u) --q;
    }
  }

  tmp = distance;         // Save
  s.assign(dimz(),1);     // Center of the upper envelope parabolas
  t.assign(s);            // Seperating index between 2 upper envelope parabolas

  cimg_forXY(*this,x,y) {
    q = s(0) = t(0) = 0;

    // Forward Scan
    for (int u=1; u<dimz(); ++u) {
      while ((q>=0) && (_saito::F(t(q),s(q),tmp(x,y,s(q)),sZ)>_saito::F(t(q),u,tmp(x,y,u),sZ))) --q;
      if (q<0) { q = 0; s(0) = u; }
      else {
        w = 1 + _saito::SEP(s(q),u,tmp(x,y,s(q)),tmp(x,y,u),sZ);
        if (w<depth) { s(++q) = u; t(q) = w; }
      }
    }

    //Backward Scan
    for (int u=dimz()-1; u>=0; --u) {
      distance(x,y,u) = _saito::F(u,s(q),tmp(x,y,s(q)),sZ);
      if (t(q)==(unsigned long)u) --q;
    }
  }

  return distance;
}

/**
 * In place version of get_distance_saito().
 */
CImg<T>& distance_saito() {
  return get_distance_saito().transfer_to(*this);
}

#endif
